package be.iminds.ilabt.jfed.experimenter_gui.dialogs;

import be.iminds.ilabt.jfed.experimenter_gui.slice.Experiment;
import be.iminds.ilabt.jfed.experimenter_gui.slice.ExperimentController;
import be.iminds.ilabt.jfed.experimenter_gui.slice.tasks.ExperimentTasksFactory;
import be.iminds.ilabt.jfed.experimenter_gui.util.DateTimeUtils;
import be.iminds.ilabt.jfed.experimenter_gui.util.ui.TransformingObjectBinding;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.highlevel.model.Sliver;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.rspec.model.GeantTestbedType;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXModelRspec;
import be.iminds.ilabt.jfed.ui.javafx.FXMLUtil;
import be.iminds.ilabt.jfed.ui.javafx.GlyphUtils;
import be.iminds.ilabt.jfed.ui.javafx.util.InformationLabel;
import be.iminds.jfed.gts_highlevel.jobs.GtsJobFactory;
import be.iminds.jfed.gts_highlevel.model.GtsReservation;
import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;
import javafx.beans.binding.DoubleBinding;
import javafx.beans.binding.ObjectExpression;
import javafx.beans.property.*;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.fxml.FXML;
import javafx.scene.Node;
import javafx.scene.control.*;
import javafx.scene.control.cell.CheckBoxTableCell;
import javafx.scene.layout.VBox;
import javafx.scene.paint.Color;
import javafx.stage.Window;
import org.controlsfx.glyphfont.FontAwesome;

import java.time.Instant;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Map;
import java.util.concurrent.FutureTask;
import java.util.stream.Collectors;

/**
 * User: twalcari
 * Date: 2/24/14
 * Time: 4:03 PM
 */

public class TerminateSliceDialog extends VBox {
    private final Window parentWindow;

    private final Dialog<Collection<TerminableExperimentPart>> dialog;
    private final Slice slice;
    private final Experiment experiment;
    private final ExperimentController experimentController;
    private final ExperimentTasksFactory experimentTasksFactory;
    private final GtsJobFactory gtsJobFactory;

    private final IntegerProperty selectedAuthoritesCount = new SimpleIntegerProperty();

    @FXML
    private TableView<TerminableExperimentPart> authoritiesTableView;

    @FXML
    private TableColumn<TerminableExperimentPart, Boolean> selectionTableColumn;
    @FXML
    private TableColumn<TerminableExperimentPart, String> nameTableColumn;
    @FXML
    private TableColumn<TerminableExperimentPart, Instant> expirationTableColumn;

    @FXML
    private InformationLabel sliceTerminationLabel;

    private ButtonType releaseResourcesButtonType = new ButtonType("Release resources", ButtonBar.ButtonData.OK_DONE);

    private final ObservableList<TerminableExperimentPart> terminateableResources;


    TerminateSliceDialog(Window parentWindow, ExperimentController experimentController, ExperimentTasksFactory experimentTasksFactory, GtsJobFactory gtsJobFactory) {
        this.parentWindow = parentWindow;

        this.experimentController = experimentController;
        this.gtsJobFactory = gtsJobFactory;
        this.experiment = experimentController.getExperiment();
        this.slice = experiment.getSlice();
        this.experimentTasksFactory = experimentTasksFactory;


        terminateableResources = FXCollections.observableArrayList();

        Multimap<SfaAuthority, Sliver> sliverPerAuthorityMap = HashMultimap.create();
        for (Sliver sliver : slice.getSlivers()) {
            sliverPerAuthorityMap.put(sliver.getAuthority(), sliver);
        }

        for (SfaAuthority authority : sliverPerAuthorityMap.keySet()) {
            Sliver firstSliver = sliverPerAuthorityMap.get(authority).iterator().next();

            terminateableResources.add(new TerminableSliversInAuthorityPart(authority,
                    sliverPerAuthorityMap.get(authority), firstSliver.expirationDateProperty()));
        }


        if (experiment.getNewRequestRspecSource() != null) {
            FXModelRspec rspec =
                    (FXModelRspec) experiment.getNewRequestRspecSource().getModelRspec(ModelRspecType.FX);

            if (rspec != null) {
                for (Map.Entry<String, GtsReservation> entry : experiment.getGtsReservations().entrySet()) {
                    GeantTestbedType gtt = rspec.getGeantTestbedTypeByName(entry.getKey());

                    terminateableResources.add(new TerminableGtsReservationPart(gtt, entry.getValue()));
                }
            }
        }

        terminateableResources.sort(Comparator.comparing(TerminableExperimentPart::getName));

        FXMLUtil.injectFXML(this);

        DialogPane dialogPane = new DialogPane();
        dialogPane.setHeaderText("Terminate Experiment");
        dialogPane.setGraphic(GlyphUtils.createDialogGlyph(FontAwesome.Glyph.STOP, Color.RED));
        dialogPane.setContent(this);

        dialogPane.getButtonTypes().setAll(releaseResourcesButtonType, ButtonType.CANCEL);

        dialog = new Dialog<>();
        dialog.initOwner(parentWindow);
        dialog.setDialogPane(dialogPane);
        dialog.setTitle("Terminate Experiment");

        dialog.setResultConverter(buttonType -> {
            if (buttonType == ButtonType.CANCEL)
                return null;
            else {
                return terminateableResources.stream()
                        .filter(TerminableExperimentPart::getTerminateRequested)
                        .collect(Collectors.toSet());
            }
        });

        //initialize dialog
        Node terminateButton = dialog.getDialogPane().lookupButton(releaseResourcesButtonType);
        terminateButton.disableProperty().bind(selectedAuthoritesCount.isEqualTo(0));
    }

    @FXML
    private void initialize() {
        //populate table
        DoubleBinding dividableWidth =
                authoritiesTableView.widthProperty().subtract(selectionTableColumn.widthProperty()).subtract(2);
        nameTableColumn.prefWidthProperty().bind(dividableWidth.multiply(0.6));
        expirationTableColumn.prefWidthProperty().bind(dividableWidth.multiply(0.4));

        authoritiesTableView.setItems(terminateableResources);
        authoritiesTableView.setPrefHeight(26/*header*/ + terminateableResources.size() * 25);

        selectionTableColumn.setCellValueFactory(t -> t.getValue().terminateRequestedProperty());
        selectionTableColumn.setCellFactory(CheckBoxTableCell.forTableColumn(selectionTableColumn));

        nameTableColumn.setCellValueFactory(t -> new ReadOnlyStringWrapper(t.getValue().getName()));

        expirationTableColumn.setCellValueFactory(t -> t.getValue().expirationTimeProperty());
        expirationTableColumn.setCellFactory(t -> new TableCell<TerminableExperimentPart, Instant>() {
            @Override
            protected void updateItem(Instant date, boolean empty) {
                if (date == null || empty) {
                    setText(null);
                } else {
                    setText(DateTimeUtils.DATE_HOUR_MINUTES_SECONDS_FORMATTER.format(LocalDateTime.ofInstant(date, ZoneId.systemDefault())));
                }
            }
        });

        assert slice.getExpirationDate() != null;

        sliceTerminationLabel.setText(String.format("While all resources can be released from your experiment '%s', the experiment container itself (slice) must expire to be invalidated. This will happen at %s.",
                slice.getName(), DateTimeUtils.DATE_HOUR_MINUTES_FORMATTER.format(LocalDateTime.ofInstant(slice.getExpirationDate(), ZoneId.systemDefault()))));
    }


    public void showDialog() {
        Collection<TerminableExperimentPart> partsToTerminate = dialog.showAndWait().orElse(Collections.emptyList());

        for (TerminableExperimentPart terminableExperimentPart : partsToTerminate) {
            if (terminableExperimentPart instanceof TerminableSliversInAuthorityPart) {
                experimentController.submitExperimentTask(
                        experimentTasksFactory.createStopExperimentAtAuthorityTask(
                                experimentController.getExperiment(), ((TerminableSliversInAuthorityPart) terminableExperimentPart).getAuthority()));
            } else if (terminableExperimentPart instanceof TerminableGtsReservationPart) {
                TerminableGtsReservationPart terminableGtsReservationPart =
                        (TerminableGtsReservationPart) terminableExperimentPart;
                new Thread(new FutureTask<>(
                        gtsJobFactory.createGtsStopTestbedJob(terminableGtsReservationPart.gtt, terminableGtsReservationPart.reservation)))
                        .start();

            } else {
                throw new IllegalArgumentException("Unknown TerminableExperimentPart");
            }
        }

    }

    private class TerminableSliversInAuthorityPart extends TerminableExperimentPart {
        private final SfaAuthority authority;
        private final Collection<Sliver> slivers;
        private final ObjectProperty<Instant> expirationTime;


        private TerminableSliversInAuthorityPart(SfaAuthority authority, Collection<Sliver> slivers, ObjectProperty<Instant> expirationTime) {
            this.authority = authority;
            this.slivers = slivers;
            this.expirationTime = expirationTime;
        }

        public SfaAuthority getAuthority() {
            return authority;
        }

        public Collection<Sliver> getSlivers() {
            return slivers;
        }

        @Override
        String getName() {
            return authority.getName();
        }

        @Override
        public Instant getExpirationTime() {
            return expirationTime.get();
        }

        @Override
        public ObjectProperty<Instant> expirationTimeProperty() {
            return expirationTime;
        }
    }

    private class TerminableGtsReservationPart extends TerminableExperimentPart {
        private final GeantTestbedType gtt;
        private final GtsReservation reservation;
        private final ObjectExpression<Instant> expirationTime;

        private TerminableGtsReservationPart(GeantTestbedType gtt, GtsReservation reservation) {
            this.gtt = gtt;
            this.reservation = reservation;

            this.expirationTime =
                    new TransformingObjectBinding<>(reservation.endTimeProperty(), ZonedDateTime::toInstant);
        }

        @Override
        String getName() {
            return gtt.getName();
        }

        @Override
        public Instant getExpirationTime() {
            return expirationTime.get();
        }

        @Override
        public ObjectExpression<Instant> expirationTimeProperty() {
            return expirationTime;
        }

    }

    private abstract class TerminableExperimentPart {

        private final BooleanProperty terminateRequested = new SimpleBooleanProperty(true);

        {
            selectedAuthoritesCount.setValue(selectedAuthoritesCount.getValue() + 1);
            terminateRequested.addListener(observable -> {
                if (terminateRequested.get()) {
                    selectedAuthoritesCount.setValue(selectedAuthoritesCount.getValue() + 1);
                } else {
                    selectedAuthoritesCount.setValue(selectedAuthoritesCount.getValue() - 1);
                }
            });
        }

        public boolean getTerminateRequested() {
            return terminateRequested.get();
        }

        public BooleanProperty terminateRequestedProperty() {
            return terminateRequested;
        }

        public void setTerminateRequested(boolean terminateRequested) {
            this.terminateRequested.set(terminateRequested);
        }

        abstract String getName();

        abstract Instant getExpirationTime();

        abstract ObjectExpression<Instant> expirationTimeProperty();
    }
}
